from typing import List


class MerchantTypeMatcher:
    def __init__(self, merchant_type: str, keywords: List[str]) -> None:
        self._keywords = keywords
        self._merchant_type = merchant_type

    @property
    def merchant_type(self):
        return self._merchant_type

    def is_match(self, description: str) -> bool:
        for keyword in self._keywords:
            if keyword.lower() in description.lower():
                return True
        return False


def determine_merchant_type(description: str) -> str:
    matchers = [
        MerchantTypeMatcher('groceries', ['new world', 'countdown', 'millwater superette', 'four square', 'Joy Mart']),
        MerchantTypeMatcher('petrol', ['Mobil Red Beach']),
        MerchantTypeMatcher('retail', ['Warehouse', 'ALIEXPRESS']),
        MerchantTypeMatcher('eatingout', ['Hollywood', 'Liquor Spot', 'Pizza Hut', 'Subway', 'China Castle', 'Olivers Cafe', 'Nam Nam', 'Mojo', 'Moreish', 'THE ISLAND GELATO', 'Tabak', 'Montana Catering', 'Starbucks', 'Sals Pizza', 'RED BEACH BAKEHOUSE', 'French Rendez-Vous', 'Wendys', 'Don Kebab', 'St Pierre', 'Devon On The Wharf', 'NO NA BAKERY', 'The Gourmet Food', 'Super Liquor', 'Hanarum', 'The Bakehouse Cafe'])
    ]
    matches = [matcher for matcher in matchers if matcher.is_match(description)]
    if len(matches) > 1:
        raise LookupError(f'too many matchers for {description} - found {len(matchers)}')
    elif len(matches) == 1:
        return matches[0].merchant_type
    return 'unknown'



from wdig.database import Transaction
from fnmatch import fnmatch


def map_to_donations_budget(tran: Transaction) -> None:
    budget_tag = None

    for m in [
        'Salvation Army NZ*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'salvation army'

    for m in [
        'PAYPAL *KIVA.ORG*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'kiva'

    for m in [
        '*OREWA BAPTIST CHURCH*',
        'Orewa Baptist Church*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'church'

    for m in [
        'AUCKLAND CITY MISSION*'
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'city mission'

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'donations'
        return True



from wdig.database import Transaction


def map_to_income_budget(tran: Transaction) -> None:
    budget_tag = None

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'income'
        return True



from typing import List
from sqlalchemy import create_engine
from sqlalchemy.exc import ProgrammingError
from sqlalchemy.orm import sessionmaker, Session, Query
from sqlalchemy.dialects.postgresql import UUID
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Date, JSON, NUMERIC, Boolean
import wdig.config as config
from wdig.parse import ParsedTransaction


_db_connection_string_no_db = f"postgresql://{config.db_username}:{config.db_password}@{config.db_hostname}:{config.db_hostport}"
_db_connection_string_with_db = _db_connection_string_no_db + f'/{config.db_dbname}'

_db_app_engine = None
_db_app_session = None


class DatabaseAppSession:
    """manages a database connection and session for the lifetime of the application"""

    @property
    def db_engine(self):
        global _db_app_engine
        if not _db_app_engine:
            _db_app_engine = create_engine(_db_connection_string_with_db, echo=False)
        return _db_app_engine

    @property
    def db_session(self) -> Session:
        global _db_app_session
        if not _db_app_session:
            session_maker = sessionmaker(bind=self.db_engine)
            _db_app_session = session_maker()
        return _db_app_session


def add_parsed_transaction(parsed_tran: ParsedTransaction) -> None:
    db_conn = DatabaseAppSession()
    tran = Transaction()

    tran.tran_id = parsed_tran.unique_id
    tran.tran_type = parsed_tran.tran_type
    tran.amount = parsed_tran.amount
    tran.tran_date = parsed_tran.tran_date
    tran.description = parsed_tran.description
    tran.account_number = parsed_tran.account_number
    tran.account_name = parsed_tran.account_name
    tran.bank_name = parsed_tran.bank_name
    tran.account_format = parsed_tran.account_format
    tran.original_columns = parsed_tran.original
    tran.imported_date = parsed_tran.imported_date
    tran.file_name = parsed_tran.file_name
    tran.is_duplicate = False

    db_conn.db_session.add(tran)


def commit_changes() -> None:
    db_conn = DatabaseAppSession()
    db_conn.db_session.commit()


def setup_database():
    engine = create_engine(
        _db_connection_string_no_db, echo=True, isolation_level="AUTOCOMMIT"
    )
    # https://docs.sqlalchemy.org/en/13/core/connections.html
    with engine.connect() as conn:
        try:
            conn.execute(f"CREATE DATABASE {config.db_dbname};")
        except ProgrammingError:
            pass

    engine = create_engine(
        _db_connection_string_with_db, echo=True, isolation_level="AUTOCOMMIT"
    )
    Base.metadata.create_all(engine)


Base = declarative_base()


class GoodBudgetTransaction(Base):
    __tablename__ = 'goodbudget_transaction'
    id = Column(Integer, primary_key=True)
    tran_date = Column(Date, nullable=True)
    envelope = Column(String(1000), nullable=True)
    details = Column(String(1000), nullable=True)
    name = Column(String(1000), nullable=True)
    check_num = Column(String(1000), nullable=True)
    notes = Column(String(1000), nullable=True)
    amount = Column(NUMERIC(12, 2), nullable=True)
    matched_tran_id = Column(UUID, nullable=True)


class Transaction(Base):
    __tablename__ = "transaction"
    id = Column(Integer, primary_key=True)
    tran_id = Column(UUID, nullable=False)
    tran_type = Column(String(16), nullable=False)
    amount = Column(NUMERIC(12, 2), nullable=False)
    tran_date = Column(Date, nullable=False)
    description = Column(String(1000), nullable=False)
    account_number = Column(String(100), nullable=False)
    account_name = Column(String(100), nullable=False)
    bank_name = Column(String(100), nullable=False)
    account_format = Column(String(100), nullable=False)
    original_columns = Column(JSON, nullable=False)
    tag = Column(String(20))
    category = Column(String(40))
    imported_date = Column(DateTime, nullable=False)
    file_name = Column(String(100), nullable=False)
    is_duplicate = Column(Boolean, nullable=False)
    merchant_type = Column(String(100), nullable=True)
    period_year = Column(String(4), nullable=True)
    period_month = Column(String(6), nullable=True)
    budget_category = Column(String(40), nullable=True)
    budget_tag = Column(String(40), nullable=True)


class TransactionIterator(DatabaseAppSession):
    def __init__(self, batch_size=10, auto_commit=True, period_month=None, period_year=None) -> None:
        super().__init__()
        self.batch_size = batch_size
        self.auto_commit = auto_commit
        self.period_month = period_month
        self.period_year = period_year

    def __iter__(self):
        self.position = 0
        return self

    def _build_base_query(self) -> Query:

        if self.db_session.dirty:
            self.commit()

        query = self.db_session.query(Transaction)
        if self.period_month:
            query = (
                query
                .filter(Transaction.period_month == self.period_month)
                .filter(Transaction.period_year == self.period_year)
            )

        return query

    def __next__(self) -> List[Transaction]:
        if self.db_session.dirty:
            self.commit()

        query = self._build_base_query()

        results = self._build_base_query() \
            .order_by(Transaction.id) \
            .offset(self.position) \
            .limit(self.batch_size) \
            .all()

        # result = self.db_session.execute(stmt).scalars().all()
        if len(results) == 0:
            raise StopIteration
        self.position += self.batch_size
        return results

    @property
    def dirty(self) -> bool:
        return self.db_session.dirty

    def commit(self) -> None:
        self.db_session.commit()

    def rollback(self) -> None:
        self.db_session.rollback()

    @property
    def total_batches(self) -> int:
        count = self._build_base_query().count()
        return count / self.batch_size


if __name__ == "__main__":
    for batch in TransactionIterator():
        print('---')
        for tran in batch:
            tran.account_name = tran.account_name
            print(tran.id)



from sqlalchemy import true
from wdig.database import Transaction
from fnmatch import fnmatch


"""
map transactions to budget items

flag transactions as "not matched" if not mapped to budget

list of rules? we can infer category by mapping budget item to category.

are we just mapping merchants to budget items? no - we don't have to map everything.

SUMMARY SPREADSHEET:

category, budget item, amount
-----------------------------
period, fixed, school fees, $amount, $budget,
fixed, power, $xxx
unknown, unknown, $xxx


car tax	$11.67
car wof & maint	$50.00
tkd oong bus	$23.33

"""


def map_to_fixed_budget(tran: Transaction) -> None:
    budget_tag = None

    if 'orcon' in tran.description.lower():
        budget_tag = 'orcon'

    if 'southern cross' in tran.description.lower():
        budget_tag = 'southern cross'

    if 'auckland council' in tran.description.lower():
        budget_tag = 'city council'

    if 'fidelity life' in tran.description.lower():
        budget_tag = 'life insurance'

    for m in [
        "*enviro waste*",
        "*ENVIROWASTE*",
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'recycling'

    for m in [
        "*PAY King's way primary school*",
        "*PAY Kingsway*",
        '*NZ CHRISTIAN PROPRIE*',
        'NZCPT kingsway*',
        'Kingsway Fees James JEONG-BURNS',
        'NZCPT*',
        '*CHRISTIAN PROPRIETORS*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'school fees'

    for m in [
        'PAY kingsway special donation acc ;',
        'Kingsway SC James JEONG-BURNS',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'school special character'

    if 'CONTACT ENERGY' in tran.description.upper():
        budget_tag = 'power'

    if 'watercare' in tran.description.lower():
        budget_tag = 'water'

    if 'aa insurance' in tran.description.lower():
        if tran.amount < -100:
            budget_tag = 'home insurance'
        else:
            budget_tag = 'car insurance'

    if 'lastpass.com' in tran.description.lower():
        budget_tag = 'password vault'

    for m in [
        "Microsoft*Realms*",
        "*MICROSOFT*REALMS*",
        "*MOJANG*",
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'minecraft realm'

    if 'apple.com' in tran.description.lower():
        budget_tag = 'apple'

    for m in [
        "GOOGLE*Google Stor*",
        "GOOGLE*YouTubePrem*",
        'Google*Storage*',
        'GOOGLE YouTube*',
        'Google YouTubePremiu*',
        'YouTubePremium*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'google storage+youtube'

    for m in [
        "*PAY Red Beach playcentre*",
        "*RBLC*",
        "PAY Orewa beach kindy*",
        'orewa beach kindy James JB',
        'Orewa Kindy James JB',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'kindy'

    if 'AMAZON MUSIC UNLIMIT' in tran.description.upper():
        budget_tag = 'amazon music'

    if 'vodafone prepay' in tran.description.lower():
        budget_tag = 'mobile usage'

    if 'snap fitness' in tran.description.lower():
        budget_tag = 'casey gym'

    if 'crashplan' in tran.description.lower():
        budget_tag = 'backup'

    if 'disney plus' in tran.description.lower():
        budget_tag = 'disney'

    if 'blake elearning' in tran.description.lower():
        budget_tag = 'reading eggs'

    for m in [
        'Fixed Expenses Taekwondo*',
        'daehan tkd*',
        'Daehan TKD*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'tkd'

    if 'microsoft*store' in tran.description.lower():
        budget_tag = 'onedrive'
    if 'microsoft*onedrive' in tran.description.lower():
        budget_tag = 'onedrive'

    if 'northern arena' in tran.description.lower():
        budget_tag = 'swim lessons'

    if 'netflix' in tran.description.lower():
        budget_tag = 'netflix'

    if 'netflix' in tran.description.lower():
        budget_tag = 'netflix'
# EA *ORIGIN.COM help.ea.com ;

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'fixed'
        return True


# PAY Daehan tkd ;



from wdig.database import Transaction


def map_to_debt_budget(tran: Transaction) -> None:
    budget_tag = None

    if 'jebu coop' in tran.description.lower():
        budget_tag = 'cc payments'

    if 'ACCOUNT FEE' in tran.description.upper():
        budget_tag = 'bank fees'

    if 'DD DISHONOUR FEE' in tran.description.upper():
        budget_tag = 'bank fees'

    if 'DRAFT INTEREST' in tran.description.upper():
        budget_tag = 'bank fees'

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'debt'
        return True



from sqlalchemy import create_engine
from sqlalchemy.orm import session, sessionmaker, declarative_base
from sqlalchemy import Column, Integer, String, DateTime, Date, JSON, NUMERIC, Boolean
from wdig.config import config
from decimal import Decimal as D
import sqlalchemy.types as types


_engine = create_engine(f"sqlite:///{config.path_to_sqlite_db}")
_session_maker = sessionmaker(bind=_engine)

def database_session():
    return _session_maker()


class SqliteNumeric(types.TypeDecorator):
    impl = types.String

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(types.VARCHAR(100))

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return D(value)


Base = declarative_base()


class Transaction(Base):
    __tablename__ = "transaction"
    id = Column(Integer, primary_key=True)
    tran_id = Column(String(16), nullable=False)
    tran_type = Column(String(16), nullable=False)
    amount = Column(SqliteNumeric(12, 2), nullable=False)
    tran_date = Column(Date, nullable=False)
    description = Column(String(1000), nullable=False)
    account_number = Column(String(100), nullable=False)
    account_name = Column(String(100), nullable=False)
    bank_name = Column(String(100), nullable=False)
    account_format = Column(String(100), nullable=False)
    original_columns = Column(JSON, nullable=False)
    tag = Column(String(20))
    category = Column(String(40))
    imported_date = Column(DateTime, nullable=False)
    file_name = Column(String(100), nullable=False)
    is_duplicate = Column(Boolean, nullable=False)
    period_year = Column(String(4), nullable=True)
    period_month = Column(String(6), nullable=True)
    budget_category = Column(String(40), nullable=True)
    budget_tag = Column(String(40), nullable=True)


def create_schema() -> None:
    Base.metadata.create_all(_engine)


def destroy_database() -> None:
    Base.metadata.delete_all()


if __name__ == '__main__':
    create_schema()
    with database_session() as db:
        pass



from wdig.database import Transaction
from fnmatch import fnmatch


def map_to_variable_budget(tran: Transaction) -> None:
    budget_tag = None

    if tran.description == 'PAY Daehan tkd ;':
        return False

    for m in [
        'AT Hop *',
        'At Hop *',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'at hop'

    for m in [
        'Mobil Red Beach*',
        'Z Greville Rd*',
        'Z Silverdale*',
        'Mobil Red Beach*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'petrol'

    for m in [
        'FAMILY MARKET- OREWA*',
        'Four Square Red Beach*',
        'Imart*',
        'IMART*',
        'ALBANY VEGE*',
        'CANAAN VEGE*',
        'Chowon Butchery*',
        'CHOWON BUTCHERY*',
        'Countdown Akl*',
        'Countdown Albert*',
        'Countdown Online*',
        'Countdown Silverdale*',
        'FOUR SQUARE RED BEAC*',
        'Four Square Red Beach*',
        'Fruit World*',
        'FRUIT WORLD*',
        'Fruit World*',
        'Hanarum Mart*',
        'Imart*',
        'IMART*',
        'Joy Mart*',
        'JOY MART*',
        'MEAT PLAZA*',
        'New World*',
        'New World*',
        'New World*',
        'Pak N Save*',
        'POS W/D FRUIT WORLD*',
        'POS W/D HANYANG*',
        'POS W/D LINK VEGE*',
        'Sams Butchery*',
        'WANG FOOD MARKET*',
        'Wang Food Market*',
        'Warkworth New World*',
        'COUNTDOWN*',
        'Countdown Orewa*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'groceries'

    for m in [
        'Asahi*',
        'Bake And Brew*',
        'BAKE AND BREW*',
        'Bakers Delight*',
        'BAMCHAN-GA*',
        'Blue Elephant*',
        'Brew On Quay*',
        'Burgerfuel Customs*',
        'Cafe Brioche Albany*',
        'Cafe Third Wave*',
        'CENTREWAY DAIRY*',
        'Chargrill 101*',
        'China Castle*',
        'CHINA CASTLE*',
        'City 1 Limited*',
        'CITY 1 LIMITED*',
        'City Convenience*',
        'Coca-Cola Ep Nz*',
        'Custom Mart*',
        'Daruma Sushi*',
        'Dear Coasties*',
        'Devon On The Wharf*',
        'Don Kebab Nz Ltd*',
        'DORURI*',
        'Eat India*',
        'Faridas*',
        'Gloria Jean*',
        'Ground Control Cafe*',
        'GROUND CONTROL CAFE*',
        'Hill Top Dairy*',
        'HOLLYWOOD BAKERY*',
        'Hollywood Bakery*',
        'House Of Boulevard*',
        "INGO'S GERMAN BAKERY*",
        'Jesters-Takapuna*',
        'Kiwi Liquor Silverdale*',
        'Ks Chicken*',
        'Leo Thai Kitchen*',
        'Lieutenant Coffee*',
        'Liquor Spot*',
        'Lord Of The Fries*',
        'Mcdonalds Albany*',
        'Mcdonalds Britomart*',
        'Millies*',
        'Millies Cafe*',
        'Millwater Fine Wines*',
        'Millwater Superette*',
        'MILLWATER SUPERETTE*',
        'Miss Moonshines*',
        'MISS MOONSHINES*',
        'Mojo Little Quay*',
        'Montana Catering*',
        'Moreish Silverdale*',
        'MR BON BAKERY*',
        "MUM'S FOODS NO.2*",
        'Nam Nam Harbour*',
        'NARA SUSHI*',
        'Natural Bake & Coffe*',
        'Nh Liquor*',
        'NO NA BAKERY*',
        'NO1 CHICKEN*',
        'Northern Union*',
        'Ocean Eats*',
        'Olivers Cafe*',
        'Omgoodness Matakana*',
        'Order Meal Ltd*',
        'OREWA BAKERY*',
        'OREWA BEACH FISH*',
        'Paraoa Brewing Co*',
        'Pizza Hut Silverdale*',
        'PIZZA HUT SILVERDALE*',
        'PIZZA HUT WHANGAPARAOA*',
        'POS W/D JAMI JAMI*',
        'POS W/D RED BEACH*',
        'POS W/D St Pierres*',
        'POS W/D WELLBEING*',
        'RED BEACH BAKEHOUSE*',
        'Robert Harris*',
        'Royal Save Mart*',
        'Sals Pizza Fort St*',
        'So French*',
        'SOPHIA DONKATSU*',
        "St Pierre'S Silverdale*",
        'STAR SHOP*',
        'Starks Cafe*',
        'Subway Orewa*',
        'Subway Parnell*',
        'Super Liquor Red Beach*',
        'Subway Orewa*',
        'Subway Parnell*',
        'Super Liquor Red Beach*',
        'THE COFFEE CLUB OREW*',
        'The Coffee Club Orewa*',
        'THE COFFEE CLUB SILV*',
        'The Gourmet Food*',
        'THE ISLAND GELATO*',
        'The White Lady*',
        'Toro Churro Botany*',
        'Wendys Whangaporara*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'eating out'

    for m in [
        'AC ALBANY STADIUM*',
        'Ac Albany Stadium*',
        'Ac Stanmore*',
        'AC STANMORE*',
        'aliexpress*',
        'ALIEXPRESS.COM*',
        'AMAZON WEB SERVICES*',
        'AMCAL SILVERDALE PHA*',
        'AMZN Digital*',
        'Dollartopia*',
        'DR SCISSORS*',
        'Briscoes Silverdale*',
        'Bunnings*',
        'Chemist Warehouse*',
        'Coast Pet Care Chari*',
        'Estuaryarts*',
        'Eventfinda*',
        'EZI*FAILED PAYMENT*',
        'Ezibuy*',
        'EZIBUY*',
        'Farmers Albany*',
        'Fullers Devonport*',
        'Furry Friends Pet*',
        'Game Over Auckland*',
        'Gmarket*',
        'Haggie Yang*',
        'Hallenstein Online*',
        'HARBOUR HOSPICE*',
        'Healthpost*',
        'HEALTHPOST*',
        'HESTIA RODNEY*',
        'Holy Trinty Cathedra*',
        'Huntfish Co New Zealan*',
        'Inflatable World*',
        'Japan Home Mart*',
        'Jeanswest*',
        'Jeanswest.Com*',
        'Jump*',
        'JUMP*',
        'Just Cuts Silverdale*',
        'Kathmandu*',
        'KINDO AUCKLAND*',
        'Kings Plant Barn*',
        'Kmart Albany*',
        'KMART ONLINE*',
        'Life Pharmacy Orewa*',
        'LOOK SHARP ALBANY*',
        'LOTSA GOODIES ALBANY*',
        'Macpac*Silverdale*',
        'Macpac Retail*',
        'MAKASSAR CORNER*',
        'Marine Deals*',
        'Mighty Ape*',
        'NEURON MOBILITY*',
        'NH MALL*',
        'Nz Transport Agency*',
        'NZ TRANSPORT AGENCY*',
        'Nzsale*',
        'NZSALE*',
        'OFFICEMAX AUCKLAND*',
        'OREWA CENTRAL POST*',
        'Orewa Central Post*',
        'PAY*',
        'Paypal *Kmart*',
        'PB TECHNOLOGIES*',
        'PHOTO LIFE STUDIOS*',
        'POS W/D AC STANMORE*',
        'POS W/D Briscoes*',
        'POS W/D Coast Pet Car*',
        'POS W/D DOLLARTOPIA*',
        'Postie - Sil*',
        'Postie Web Store*',
        'Rebel Sport Mt Albert*',
        'Rebel Sport Silverdale*',
        'Red Beach Pharmacy*',
        'SILVERDALE CLINIC*',
        'SPECSAVERS NZ*',
        "St John'S Carpark*",
        'STARZ LIVERPOOL*',
        'STEAM PURCHASE*',
        'SUSHI GALLERY*',
        'Sushi House*',
        'TCS SILVERDALE*',
        'The Civic*',
        'The Doctors Coastcare*',
        'THE DOCTORS RED BEACH*',
        'THE PUMPHOUSE THEATRE*',
        'The Warehouse*',
        'Tm *Ticketmasternz*',
        'TM *TICKETMASTERNZ*',
        'TRADEME*',
        'Trademe*',
        'Udemy*',
        'We-Life*',
        'WE-LIFE*',
        'Whitcoulls Aly*',
        'www.aliexpress.com*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'entertainment & misc'

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'variable'
        return True


"""
# ----------------------
# what are these?
# JEBI
# jebu kiwibank rev
#  --------------------------------

# Google Auckland ;
# Google Auckland ;
# Google Google          Auckland      Nz
# GOOGLE Google Auckland ;
# GOOGLE Warner Bros Auckland ;

# Rodneykids Ltd         Auckland      Nz
# kingsway junior spor 3ang- ethan
# Kingsway School        Silverdale    Nz
# kingsway junior spor 3ang- ethan
# Kingsway School        Silverdale    Nz

# POS W/D 14.95AUD @ 0.9320 conversion rate ;(INC. $0.40 CURRENCY CONVERSION COMMISSION) Audible Limited AU MELBOURNE
# POS W/D 16.45AUD @ 0.9109 conversion rate ;(INC. $0.45 CURRENCY CONVERSION COMMISSION) Audible Limited AU MELBOURNE
# Audible Limited Au     Melbourne     Au
"""



from wdig.database import TransactionIterator, Transaction
from wdig.enrich.budget_mapping_debt import map_to_debt_budget
from wdig.enrich.budget_mapping_fixed import map_to_fixed_budget
from wdig.enrich.budget_mapping_variable import map_to_variable_budget
from wdig.enrich.budget_mapping_mortgage import map_to_mortgage_budget
from wdig.enrich.budget_mapping_savings import map_to_savings_budget
from wdig.enrich.budget_mapping_donations import map_to_donations_budget
from wdig.enrich.budget_mapping_income import map_to_income_budget
from wdig.enrich.budget_mapping_transfer import map_to_transfer_budget
from rich.progress import track
import logging


def map_to_budget(tran: Transaction):
    if map_to_fixed_budget(tran):
        return
    if map_to_debt_budget(tran):
        return
    if map_to_variable_budget(tran):
        return
    if map_to_mortgage_budget(tran):
        return
    if map_to_savings_budget(tran):
        return
    if map_to_donations_budget(tran):
        return
    if map_to_income_budget(tran):
        return
    if map_to_transfer_budget(tran):
        return


# TODO make this multi-threaded
if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
    logging.info('hi!')

    period_month = '10-Oct'
    period_year = '2022'

    interator = TransactionIterator(period_month=period_month, period_year=period_year)
    for batch in track(interator, total=interator.total_batches, description='budget mapping transactions...'):
        for tran in batch:
            map_to_budget(tran)



import click
from wdig.config import config
from rich.console import Console
from rich.prompt import Prompt
from wdig.database_v2 import create_schema


console = Console()


@click.group('config')
def cli_config():
    pass

@click.command()
def check():
    config_values = config.to_dict()
    console.print(f'read config file [bold green]\u2713')

@click.command()
def show():
    console.print(f'config stored in: {config.path_to_config_file}')
    console.print(config.to_toml())

@click.command()
def reset():
    config.reset_to_default()

@click.command()
def setup():
    bank_file_path = Prompt.ask("Enter path to bank files", default=config.path_to_bank_files)

    config.path_to_bank_files = bank_file_path
    console.print(f'update config [bold green]\u2713')

    create_schema()
    console.print(f'create db schema [bold green]\u2713')


cli_config.add_command(check)
cli_config.add_command(show)
cli_config.add_command(reset)
cli_config.add_command(setup)



from dataclasses import dataclass
from datetime import datetime
import hashlib, re
from typing import Pattern


@dataclass(init=False)
class ParsedTransaction():
    # class with structure and validation!
    # transaction datamodel!!!!
    # create from dict + validation?

    def __init__(
        self,
        original: str,
        file_name: str
    ) -> None:
        self.original = original
        self.file_name = file_name
        self.imported_date = datetime.now().date()

    def to_dict(self):
        vars(self)

    unique_id: str
    tran_type: str  # credit or credit
    amount: float
    tran_date: datetime
    description: str
    account_number: str
    account_name: str
    bank_name: str
    account_format: str
    original: dict
    imported_date: datetime
    file_name: str


@dataclass
class BankAccount():
    account_name: str
    account_number: str
    bank_name: str
    account_file_format: str
    file_name_match_regex: Pattern

    def is_match(this, file_name: str) -> bool:
        return bool(re.search(this.file_name_match_regex, file_name))

    def parse_row(this, row, file_name: str) -> ParsedTransaction:
        tran = ParsedTransaction(
            file_name=file_name,
            original=row
        )
        tran.account_format = this.account_file_format
        tran.account_name = this.account_name
        tran.account_number = this.account_number
        tran.bank_name = this.bank_name
        if this.account_file_format == 'bnz-csv':
            this._parse_row_bnz(row, tran)
        elif this.account_file_format == 'kiwibank-csv':
            this._parse_row_kiwibank_cash(row, tran)
        elif this.account_file_format == 'anz-cc-csv':
            this._parse_row_anz_creditcard(row, tran)
        elif this.account_file_format == 'anz-csv':
            this._parse_row_anz_cash(row, tran)
        else:
            raise LookupError(f'Unknown bank file format {this.account_file_format}')
        return tran

    def _parse_row_bnz(this, row, tran: ParsedTransaction) -> None:
        tran.unique_id = hashlib.md5(bytes(f"{row['Payee']}-{row['Particulars']}-{row['Amount']}-{row['Date']}", encoding="utf-8",)).hexdigest()
        tran.description = f"{row['Payee']} {row['Particulars']}"
        tran.amount = float(row["Amount"])
        tran.tran_date = datetime.strptime(row["Date"], "%d/%m/%y").date()
        tran.tran_type = "debit"

    def _parse_row_kiwibank_cash(this, row, tran) -> ParsedTransaction:
        tran.unique_id = hashlib.md5(bytes(f"{row['Memo/Description']}-{row['Amount']}-{row['Date']}", encoding="utf-8")).hexdigest()
        tran.description = row["Memo/Description"]
        tran.amount = float(row["Amount"])
        tran.tran_date = datetime.strptime(row["Date"], "%d-%m-%Y")
        tran.type = "debit"
        if row["Amount (credit)"] is not None:
            tran.tran_type = "credit"

    def _parse_row_anz_creditcard(this, row, tran: ParsedTransaction) -> None:
        tran.unique_id = hashlib.md5(bytes(f"{row['Type']}-{row['Amount']}-{row['TransactionDate']}", encoding="utf-8", )).hexdigest()
        tran.description = row["Details"]
        tran.tran_date = datetime.strptime(row["TransactionDate"], "%d/%m/%Y")
        tran.amount = float(row["Amount"])
        tran.tran_type = "debit"
        if row["Type"] == "C":
            tran.tran_type = "credit"
        if tran.tran_type == "debit":
            tran.amount = tran.amount * -1

    def _parse_row_anz_cash(this, row, tran: ParsedTransaction) -> None:
        tran.unique_id = hashlib.md5(bytes(f"{row['Type']}-{row['Amount']}-{row['Date']}", encoding="utf-8",)).hexdigest()
        tran.description = f"{row['Details']} {row['Particulars']} {row['Code']}"
        tran.amount = float(row["Amount"])
        tran.tran_date = datetime.strptime(row["Date"], "%d/%m/%Y")
        tran.tran_type = "debit"
        if tran.amount < 0:
            tran.tran_type = "credit"


class BankAccountLookup():
    bank_accounts = [
        BankAccount('BNZ Cash', '02-1265-0038877-001', 'BNZ', 'bnz-csv', r'^Cash-'),
        BankAccount('BNZ Fixed Expenses', '02-1265-0038877-000', 'BNZ', 'bnz-csv', r'^Fixed-Expenses-'),
        BankAccount('BNZ Revolving Mortgage', '02-1265-0038877-083', 'BNZ', 'bnz-csv', r'^Revolving-Mortgage-'),
        BankAccount('Kiwibank Online Bills', '38-9010-0821604-08', 'Kiwibank', 'kiwibank-csv', r'38-9010-0821604-08'),
        BankAccount('Kiwibank Savings', '38-9010-0821604-16', 'Kiwibank', 'kiwibank-csv', r'38-9010-0821604-16'),
        BankAccount('Kiwibank Cash', '38-9010-0821604-09', 'Kiwibank', 'kiwibank-csv', r'38-9010-0821604-09'),
        BankAccount('ANZ Visa Gold', '4367-xxxx-xxxx-7631', 'ANZ', 'anz-cc-csv', r'4367-xxxx-xxxx-7631'),
        BankAccount('ANZ Cash', '06-0229-0443344-00', 'ANZ', 'anz-csv', r'06-0229-0443344-00'),
    ]

    def find_from_file_name(this, file_name: str) -> BankAccount:
        matches = [account for account in this. bank_accounts if account.is_match(file_name)]
        if len(matches) == 0:
            raise LookupError(f'unknown bank account for file {file_name}')
        elif len(matches) > 1:
            raise LookupError(f'file {file_name} has {len(matches)} matched bank accounts')
        return matches[0]


# TODO #51 why is this failing when built in container? param: row: dict[str, any]
def parse_row(row, file_name: str) -> ParsedTransaction:
    account_lookup = BankAccountLookup()
    account = account_lookup.find_from_file_name(file_name)
    return account.parse_row(row, file_name)



"""
transactions are allocated to periods, periods start around payday and can track spending until next pay day

example period: Aug 18th to Sept 17th -> `Aug 21`

plan is - support controlling spending by allocating budget amount to accounts on payday

what is our per-period spending per account?

"""

from datetime import datetime, timedelta
from wdig.database import Transaction


class Period():
    def __init__(self, date: datetime) -> None:
        # jan 1 > 15 --> dec peroid
        # jan 16 > end --> jan period

        if date.day < 16:
            date = date - timedelta(days=(date.day + 1))
        self._month = date.strftime('%m-%b')
        self._year = str(date.year)

    @property
    def year(self) -> str:
        return self._year

    @property
    def month(self) -> str:
        return self._month


def apply_period(tran: Transaction):
    # NOT USING PERIODS ANY MORE!
    # period = Period(tran.tran_date)
    tran.period_month = tran.tran_date.strftime('%m-%b')
    tran.period_year = str(tran.tran_date.year)



import os
from rich.console import Console
from wdig.parse import parse_row
import csv
from wdig.database import add_parsed_transaction, commit_changes
from wdig.queries import is_file_already_loaded
import wdig.config as config
from wdig.object_storage import BankFileObjectStorage


console = Console()

class BankFileLoader():

    def load_new_files(self) -> int:

        file_loaded_count = 0
        for root, dirnames, filenames in os.walk(config.data_in_path):
            for filename in filenames:
                if not filename.lower().endswith('.csv'):
                    continue

                in_filepath = os.path.join(root, filename)
                if self._is_file_already_loaded(filename):
                    continue

                self._save_object(filename, in_filepath)

                with open(in_filepath, 'r') as infile:
                    trans = []
                    for row in csv.DictReader(infile):
                        tran = parse_row(row, filename)
                        self._add_parsed_transaction(tran)
                        trans.append(tran)
                self._commit_changes()
                file_loaded_count += 1
                console.print(f'loaded {len(trans)} in {os.path.basename(in_filepath)}')

        console.print(f'[bold green] found {file_loaded_count} new files')
        return file_loaded_count

    def _save_object(self, filename: str, path_to_file: str) -> None:
        object_storage = BankFileObjectStorage()
        object_storage.save_file(key=filename, path_to_file=path_to_file)

    def _is_file_already_loaded(self, filename):
        return is_file_already_loaded(filename)

    def _add_parsed_transaction(self, tran):
        add_parsed_transaction(tran)

    def _commit_changes(self):
        commit_changes()


# TODO the abstraction (resource access) should be testable!
class BankFileLoaderTestable(BankFileLoader):
    def __init__(self, is_file_already_loaded: bool) -> None:
        super().__init__()
        self.is_file_already_loaded = is_file_already_loaded

    def _is_file_already_loaded(self, filename):
        return self.is_file_already_loaded

    def _save_object(self, filename: str, path_to_file: str) -> None:
        pass

    def _add_parsed_transaction(self, tran):
        pass

    def _commit_changes(self):
        pass


if __name__ == '__main__':
    bfl = BankFileLoader()
    bfl.load_new_files()



"""
found this isn't easy to match -- probably dates wrong?
counts: {'matched': 51, 'duplicated': 1, 'nomatch': 138}


download goodbudget files (weekly?)
load function! -- use existing loadfunction -- need somewhere to store them (new table with forign key?).
try to match each to a transaction (new columns: goodbudget desc, goodbudget date, good budget envelope)
not matches goto unmatched spreadsheet worksheet

summary of variable by goodbuget envelope - can use transactions worksheet filtering?

TODO use a view for querying transactions - including linked GB trans

"""

from csv import DictReader
import datetime
from typing import Counter
from wdig.database import GoodBudgetTransaction, Transaction
from wdig.database import DatabaseAppSession


def load():
    gb_trans = []
    with open('data-in/goodbudget/history.csv', 'r') as f:
        for row in DictReader(f):
            gbt = GoodBudgetTransaction()
            gbt.tran_date = datetime.datetime.strptime(row['Date'], '%d/%m/%Y')
            gbt.amount = float(row['Amount'])
            gbt.check_num = row['Check #']
            gbt.details = row['Details']
            gbt.envelope = row['Envelope']
            gbt.name = row['Name']
            gbt.notes = row['Notes']
            gb_trans.append(gbt)

    db = DatabaseAppSession()
    db.db_session.query(GoodBudgetTransaction).delete()  # need to delete all and recreate - GB exports 1 file will all transactions
    db.db_session.add_all(gb_trans)
    db.db_session.commit()


def match():
    db = DatabaseAppSession()
    gb_trans = db.db_session.query(GoodBudgetTransaction).all()

    counts = {'matched': 0, 'duplicated': 0, 'nomatch': 0}
    for gb_tran in gb_trans:
        match_trans = db.db_session.query(Transaction) \
            .filter(Transaction.is_duplicate is False, Transaction.tran_date == gb_tran.tran_date.strftime('%Y-%m-%d'), Transaction.amount == gb_tran.amount) \
            .all()
        if len(match_trans) == 1:
            gb_tran.matched_tran_id = match_trans[0].tran_id
            db.db_session.commit()
            counts['matched'] += 1
        elif len(match_trans) > 1:
            counts['duplicated'] += 1
            # raise Exception('duplicate matches!')
        else:
            counts['nomatch'] += 1
    print(counts)


if __name__ == '__main__':
    load()
    match()



from fnmatch import fnmatch
from wdig.database import Transaction
from wdig.queries import get_tran_by_id


"""

variable	unknown	M Jeng 	-$1,000.00

"""


def apply_category_to_tran(tran_id: str) -> Transaction:
    tran = get_tran_by_id(tran_id)
    apply_category(tran)
    return tran


def apply_category(tran: Transaction) -> None:
    if _is_transfer(tran.description, tran.account_format, tran.original_columns.get('Type'), tran.original_columns.get('Details')):
        tran.category = 'transfer'
    elif _is_income(tran.description):
        tran.category = 'income'
    elif _is_interest(tran.description):
        tran.category = 'interest'
    elif _is_donation(tran.description):
        tran.category = 'donation'
    elif _is_mortgage(tran.description):
        tran.category = 'mortgage'
    elif _is_saving(tran.description):
        tran.category = 'savings'
    elif _is_bank_fee(tran.description):
        tran.category = 'bank fees'
    elif _is_fixed(tran.description):
        tran.category = 'fixed'
    elif _is_debt(tran.description):
        tran.category = 'debt'
    elif _is_tax(tran.description):
        tran.category = 'tax'
    else:
        tran.category = 'variable'


def _is_tax(description: str) -> bool:
    return 'INLAND REVENUE' in description


def _is_transfer(description, account_format, original_type, original_details) -> None:

    if description in [
        '4367-****-****-7631 Debit Transfer',
        'Jebu ANZ transfer',
        'Fixed Expenses fixed exp',
        'AP#20154352 TO M.JEONG ;Bnz revolvin',
        'TRANSFER TO C BURNS  M JEONG - 16 ;',
        'Cashback Visa Gold  ',
        'TRANSFER TO C BURNS  M JEONG - 08 ;',
        'TRANSFER FROM C BURNS  M JEONG - 08 ;',
        'TRANSFER FROM C BURNS  M JEONG - 09 ;',
        'TRANSFER TO C BURNS  M JEONG - 09 ;',
        'TRANSFER FROM C BURNS  M JEONG - 16 ;',
        'Revolving Mortgage',
        'CASEY & MIHEE cash',
        'Burns C Transfer ',
        'CASEY & MIHEE fixed exp',
        'C BURNS  M JEONG Bnz revolvin',
        'CASEY & MIHEE cash',
        'C Burns  M Jeong Anz Credit Acc',
        'Revolving Mortgage ',
        'Fixed Expenses ',
        'Mihee Anz Jebu ',
        'Cash cash',
        'JEBU ANZ ANZ JEBU',
        'Online       Payment -  Thank You ',
        'Automatic Payment - Thank You ',
        'Direct Credit jebu bnz MIHEE ;Ref: jebu bnz MIHEE',
        'Direct Credit JEBU A MIHEE ;Ref: JEBU A MIHEE',
        'Cash ',
        'C Burns  M Jeong From Kiwibank',
        'AP#15249025 FROM C BURNS  M JEONG ;Transfer from C BURNS  M JEONG - 09',
        '4367-****-****-7631 Credit Transfer',
        'PAY M jeong ;ANZ credit Acc',
        'AP#15249025 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 08',
        'PAY ANZ Cash Account ;To ANZ',
        'TRANSFER FROM M JEONG - 00 ;'
    ]:
        return True
    else:
        return False  # why care about transfers? they are hard to categorize. they should equal 0.

    if account_format == "anz-csv":
        if original_type == "Transfer":
            return True
        if (original_type == "Bill Payment" and original_details == "C Burns  M Jeong"):
            return True
        if (original_type == "Automatic Payment" and original_details == "Cashback Visa Gold"):
            return True

    if account_format == "kiwibank-csv":
        for p in [
            "TRANSFER FROM C BURNS M JEONG - 09*",
            "*AP#15249025 FROM C BURNS*M JEONG ;Transfer from C BURNS  M JEONG - 09*",
            "AP#14783586 TO JEBU BNZ ;jebu kb transfer*",
            "AP#15249025 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 08*",
            "TRANSFER TO C BURNS  M JEONG - 08*",
            "PAY ANZ Cash Account ;To ANZ",
            "AP#16085738 TO ANZ CASH ACCOUNT ;To ANZ*",
            "PAY ANZ Cash Account ;To ANZ*",
            "AP#16085738 TO ANZ CASH ACCOUNT ;To ANZ*",
            "PAY M jeong ;ANZ credit Acc*",
            "Bill Payment BURNS CASEY*",
            "Jebu Kiwibank*",
            "TRANSFER*BURNS*JEONG - 09*",
            "TRANSFER FROM*BURNS*JEONG*09*",
            "TRANSFER FROM*BURNS*JEONG*08*",
            "TRANSFER TO*BURNS*JEONG*16*",
            "TRANSFER FROM*BURNS*JEONG*16*",
            "PAY M JEONG*",
        ]:
            if fnmatch(description, p):
                return True

    if account_format == "anz-cc-csv":
        if original_details in [
            "Automatic Payment - Thank You ",
            "Online       Payment -  Thank You ",
        ]:
            return True

    for p in [
        'AP#20154352',  # to BNZ revolving
        'C BURNS  M JEONG Bnz revolvin',  # received bnz revolving ap
        'PAY M.jeong ;Bnz revolvin'
    ]:
        if p in description:
            return True


def _is_income(description: str) -> bool:
    for p in [
        "Salary*",
        "*FROM ONEHUNGA TOY LIBRARY*",
        "*Direct Credit FRAEDOM*",
        "*Autocashbacks*",
        "TRANSFER FROM*JEONG*00*",
        'Polyglot Group Pty*',
        'Direct Credit*I.R.D.*',
        'Jeong Chang Weon 22228180737C S000 Imt',
    ]:
        if fnmatch(description, p):
            return True


def _is_interest(description: str) -> bool:
    for p in ["*INTEREST DEBIT*", "*Interest*", "*INTEREST*"]:
        if fnmatch(description, p):
            return True


def _is_mortgage(description: str) -> bool:
    for p in [
        "*JEBU WESTPAC*",
        "*lump sum pmt*",
        "*Direct Credit from westpac Burns*",
        "*Lump Sum Payment*",
        "*TO JEBU BNZ ;jebu kb transfer*",  # AP to BNZ
        "*AP#15260870 TO C BURNS M JEONG*",  # new house mortgage
        "*AP#15260872 TO C BURNS M JEONG*",  # new house mortgage
        "AP#15260870 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 14",
        "AP#15260872 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 15",
        "AP#16085738 TO ANZ CASH ACCOUNT ;To ANZ",
        "AP#16085738 TO ANZ CASH ACCOUNT ;To ANZ",
        "MANUAL LOAN PAYMENT*",
        'CASEY & MIHEE LOAN PAYMT*',
        '*LOAN FUNDS*',
        'Discharge NA*',
        'Home Loan ',
        '1212/2120000000/002 SOLICITOR',
        'Direct Credit Refinance The Legal Team*',

    ]:
        if fnmatch(description, p):
            return True


def _is_donation(description) -> bool:
    for p in [
        "VICTORY CHURCH",
        "Salvation Army",
        "KIVA",
        "Victory church",
        "KIVA.ORG",
        "PAY NZ Victory church",
        "AUCKLAND CITY MISSION",
        "AP#18688032 TO OREWA BAPTIST CHURCH",
        'AP#20400921 TO OREWA BAPTIST CHURCH',
        'Orewa Baptist Church*',
        'AUCKLAND CITY*',
    ]:
        if p in description:
            return True


def _is_saving(description) -> bool:
    for p in [
        "*KIWI WEALTH KIWISAVER SCHEME*",
        "*OONG SAVINGS*",
        "*SAN SAVINGS*",
        "*SMARTSHARES*",
        "*Sharesies*",
        "*SHARESIES*",
        'SUPERLIFE WORKPLACE*',
        'Kiwibank savings kb savings'
    ]:
        if fnmatch(description, p):
            return True


def _is_bank_fee(description) -> bool:
    for p in [
        # '*conversion rate*', # not always fees
        "*Monthly A/C Fee*",
        "*OTHER BANK ATM FEE*",
        "*VisaDebitAnnualFee*",
        "*REPLACEMENT ATM/EFTPOS*",
        "INTEREST DEBIT*" "*Unarranged Overdraft Fee*",
        "REPLACEMENT ATM/EFTPOS*",
        "VisaDebitAnnualFee*",
        "*DISHONOUR FEE*",
        "CLOSE O'DRAFT INTEREST",
        '*ACCOUNT FEE*',
        '*Joint/Additional Account Fee*',
        'Account Fee*',
    ]:
        if fnmatch(description, p):
            return True


def _is_fixed(description) -> bool:
    for p in [
        "*Netflix.Com*",
        "*Contact Energy Ltd Wellington Nz*",
        "*Northern Arena Silverdale Nz*",
        "*ENVIRO WASTE SERVICES LIMITED*",
        "*Vodafone Prepay*",
        "*PAY Red Beach playcentre*",
        "*PAY King's way primary school*",
        "*Snap Fitness Newstead*",
        "*Direct Debit -AUCKLAND CITY COUNCIL*",
        "*PAY orewa beach kindergarten*",
        "*Direct Debit -FIDELITY LIFE ASSURANCE*",
        "*Direct Debit -SOUTHERN CROSS MEDICAL*",
        "*AP#13629838 TO ORCON INTERNET*",
        "*AMAZON MUSIC UNLIMITED SYDNEY*",
        "*PAY Global Martial Art*",
        "*Watercare Services*",
        "*Northern Arena Limit*",
        "*LIGHTBOX AUCKLAND*",
        "*Aa Insurance Auckland Nzl*",
        "*GOOGLE Music g.co/helppay*",
        "*PAY CONTACT ENERGY ;BURNSC 500365729 KIWIBANK*",
        "*Aa Insurance*",
        "*APPLE.COM/BILL SYDNEY*",
        "*MICROSOFT*REALMS PLUS MMSBILL.INFO*",
        "*NEW ZEALAND CHRISTIAN PROPRIETORS TRUST*",
        "*CONTACT ENERGY*",
        "*FIDELITY LIFE*",
        "*Watercare Online*",
        "*Ezi*Snap Fitness*",
        "*Crashplan*",
        "*MOJANG*",
        "*RBLC*",
        "*PAY WATERCARE SERVICES*",
        "*PAY Kingsway*",
        "PAY Daehan tkd*",
        "Microsoft*Realms*",
        "*Sports4Tot*",
        "PAY Joseph*",
        "Neon*",
        "Contact Energy*",
        "PAY Orewa beach kindy*",
        "Orcon*",
        'Disney PLUS*',
        'AUCKLAND COUNCIL*',
        '*NZ CHRISTIAN PROPRIE*',
        'NZCPT kingsway*',
        'Microsoft*Store*',
        '*SNAP FITNESS*',
        '*EZIDEBIT FOR AUCKLAND CITY COUNCIL*',
        'Daehan TKD Ethan JB TKD',
        'PAY kingsway special donation acc ;',
        'DRI*CrashPlan for SB 557268497805',
        'Blake eLearning NZD Leichhardt ;',
        'NETFLIX.COM 57268497805',
        'Disney Plus 57268497805',
        'AMAZON MUSIC UNLIMIT 557268497805',
        'MICROSOFT*REALMS 1 M 57268497805',
        'APPLE.COM/BILL 57268497805',
        'orewa beach kindy James JB',
        'Microsoft*OneDrive S 57268497805',
        'Southern Cross Healt ',
        'Fixed Expenses Taekwondo fe',
        'AA INSURANCE 57268497805',
        'Kingsway Fees James JEONG-BURNS',
        'Kingsway SC James JEONG-BURNS',
        'Snap Fitness Albany    Albany        Nz ',
        'Google Storage Auckland*',
        'YouTubePremium Auckland*',
        'Netflix New Zealand 57268497805',
        'Orewa Kindy James JB',
        'LastPass.com*',
        'Auckland Council*',
        'Google YouTubePremiu 57268497805',
        'NZCPT*',
    ]:
        if fnmatch(description, p):
            return True


def _is_debt(description) -> bool:
    for p in [
        'AP#19180516 TO ASB JEBU ;ASB Transfer',
        'TRF*4327*6018*',
        'AP#20181673*',  # coop CC
        'jebu coop*',
    ]:
        if fnmatch(description, p):
            return True


if __name__ == '__main__':
    tran = apply_category_to_tran('574210f1-1184-ca63-ddbd-42f0fd21323e')
    print(tran.category)
    pass



from wdig.database import Transaction
from fnmatch import fnmatch


def map_to_savings_budget(tran: Transaction) -> None:
    budget_tag = None

    for m in [
        'M Jeng*',
        'Kiwibank savings kb savings*',
        'Direct Debit -SMARTSHARES LIMITED*',
        'SUPERLIFE WORKPLACE*',
    ]:
        if fnmatch(tran.description, m):
            budget_tag = 'savings'

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'savings'
        return True



from wdig.database import Transaction


def map_to_transfer_budget(tran: Transaction) -> None:
    budget_tag = None

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'income'
        return True

# M Jeng
# AP#15260870 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 14
# PAY M.jeong ;Bnz revolvin
# AP#15260870 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 14
# AP#15260872 TO C BURNS  M JEONG ;Transfer to C BURNS  M JEONG - 15
# Revolving Mortgage Taekwondo fe



# from wdig.health_checks import HealthChecks
# from wdig.load import BankFileLoader
import click
from rich.console import Console
# from wdig.queries import query_by_period, get_most_recent_tran_date, get_tran_by_id, get_transactions_grouped_by_budget_tag
from rich.table import Table
from wdig.config import config
# from wdig.google_sheets import update_empty_sheet, update_transaction_sheet
# from wdig.database import setup_database
# from wdig.enrich import enrich_all_transactions
# from datetime import timedelta, datetime
# from wdig.object_storage import BankFileObjectStorage
from rich import inspect
import logging
import pathlib
from wdig.cli_config import cli_config
from wdig.cli_files import cli_files
from wdig.user_notifier import NotificationMessage, UserNotifier


log_path = f'{pathlib.Path(__file__).parent.resolve()}/wdig.log'
logging.basicConfig(filename=log_path,
                    filemode='a',
                    format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                    datefmt='%H:%M:%S',
                    level=logging.INFO)
logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)
logging.info('wdig starting up!')


console = Console()


def handle_user_notification(msg: NotificationMessage) -> None:
    console.print(f'{msg.message}')


notifier = UserNotifier()
notifier.subscribe(handle_user_notification)


@click.group(help=f"Where Did It Go? (wdig)")
@click.version_option()
def cli():
    pass


cli.add_command(cli_config)
cli.add_command(cli_files)
if __name__ == "__main__":
    cli()


# @click.command(help="runs health checks and tests")
# def health():
#     health_check = HealthChecks()
#     health_check.check_health()

# # todo #52 re-enrich tran command
# # wdig enrich -> enrich if missing enrichment
# # wdig enrich --all -> enrich all
# # wdig enrich --tran-id xyz -> re-enrich tran
# # wdig enrich --update -> update too?

# @click.command(help="loads new data files")
# def load():
#     bfl = BankFileLoader()
#     count = bfl.load_new_files()

#     if count == 0:
#         last_tran_date = get_most_recent_tran_date()
#         last_tran_date += timedelta(days=-2)
#         console.print(f'go and download files from "{last_tran_date.strftime("%-d %b %y")}" to today')
#         console.print('these accounts; Kiwibank (cash, savings, onlinebills) bnz (revolving, cash, fixed) anz (cash, creditcard)')
#         console.print(f'put them in a folder named {datetime.now().strftime("%Y-%m-%d")}')


# @click.command(help="Count and Amount summary by period")
# @click.option('-m', '--months-back', type=int, default=6, required=True)
# def period(months_back: int, out_data_path: str):
#     df = query_by_period(out_data_path, months_back)
#     table = Table(show_header=True, header_style="bold magenta")
#     table.add_column("period_month", style="dim", width=12)
#     table.add_column("period_year", style="dim", width=12)
#     table.add_column("count", style="dim", width=12)
#     table.add_column("amount", style="dim", width=12)
#     for row in df.values:
#         table.add_row(str(row[0]), row[1], str(row[2]), str(row[3]))
#     console.print(table)


# @click.command(help='update google sheets')
# @click.option('-y', '--year', type=str, default=2022, required=True)
# def update(year):
#     with console.status(f"[bold]Updating empty test sheet...") as status:
#         update_empty_sheet()
#         console.print('Empty test sheet updated')

#     with console.status(f"[bold]Updating transation {year} year sheet...") as status:
#         update_transaction_sheet(year)
#         console.print(f'Transation {year} year sheet updated')


# @click.command(help='creates db and schema')
# def setupdb():
#     setup_database()


# @click.command(help='enrich all transactions')
# def enrich():
#     enrich_all_transactions()

# @click.command(help='Show info about a transaction')
# @click.option('-t', '--tran_id', type=str, required=True)
# def tran(tran_id):
#     tran = get_tran_by_id(tran_id)
#     inspect(tran, methods=False)


# @click.command(name='list')
# def files_list():
#     os = BankFileObjectStorage()
#     files = os.list_files()
#     for file in files:
#         console.print(file)


# @click.command(name='budget')
# def budget():
#     budget_tags = get_transactions_grouped_by_budget_tag()

#     current_period = ''
#     for bt in budget_tags:
#         if current_period != bt.period:
#             console.print('---')
#             current_period = bt.period
#         console.print(f'{bt.period} {bt.budget_tag} {bt.amount}')

# files.add_command(files_list)
# cli.add_command(health)
# cli.add_command(load)
# cli.add_command(period)
# cli.add_command(update)
# cli.add_command(setupdb)
# cli.add_command(enrich)
# cli.add_command(tran)
# cli.add_command(budget)



"""
MS is a P&L summary for the month
 - gross income
 - expenses by categry vs budget
 - remaining

purpose is to see:
 - did we spend less then we earnt in that month? and by how much
 - are our budgets accurate

this is something that is only useful retrospectivily.  not good for tracking progress.

by period? -- calandar for now


LAYOUT

title (from to dates)
messages (i.e. transfers not zero)

table:
 income, amount, budget, remaining, %
 category, amount, budget, remaining, %
 remaining, amount, budget, remaining, %

"""


# get the data in the right structure
# pop it on a gsheet

from dataclasses import dataclass
from typing import List
from sqlalchemy import extract, func
from wdig.database import Transaction, DatabaseAppSession
from wdig.google_sheets import get_transaction_sheet, _create_or_replace_worksheet


month = 5
year = 2022

# TODO group by period_year and period_month
db = DatabaseAppSession()
trans = db.db_session.query(
        Transaction.category,
        func.sum(Transaction.amount),
        func.count(Transaction.tran_id)) \
    .filter(
        Transaction.is_duplicate == False,  # noqa E712
        extract('year', Transaction.tran_date) == year,
        extract('month', Transaction.tran_date) == month) \
    .group_by(Transaction.category) \
    .all()


# TODO read this from the budget sheet
# bank fees includes debt
budget = {'income': 9600, 'fixed': -1798.74, 'variable': -3108.33, 'mortgage': -2594.98, 'savings': -950, 'donation': -293.33, 'bank fees': -95, 'debt': -340}


@dataclass
class CategorySummary:
    category: str
    actual: float
    budget: float
    remaining: float
    percentage: int

    def as_list(self) -> List:
        return [self.category, self.actual, self.budget, self.remaining, self.percentage]

@dataclass
class Statement:
    title: str
    message: List[str]
    income: CategorySummary
    expense_summaries: List[CategorySummary]


data = Statement(
    title='Monthly Statement',
    message='',
    income=None,
    expense_summaries=[]
)


for tran in trans:
    if tran[0] == 'transfer':
        data.message = f'transfers should be zero -- {tran[1]}'
        continue

    cat_budget = budget[tran[0]]
    amount = float(tran[1])
    cs = CategorySummary(
        category=tran[0],
        actual=amount,
        budget=cat_budget,
        remaining=amount - cat_budget,
        percentage=(amount / cat_budget)
    )

    if tran[0] == 'income':
        data.income = cs
    else:
        data.expense_summaries.append(cs)

# TODO add remaining amount line


sheet = get_transaction_sheet()
worksheet = _create_or_replace_worksheet(sheet, f'{month} statement', rows=50)
worksheet.update('A1', data.title)
worksheet.update('A2', data.message)

worksheet.update('A4', [['category', 'actual', 'budget', 'remaining', '%']])
worksheet.update('A5', [data.income.as_list()])

for idx, cat in enumerate(data.expense_summaries):
    worksheet.update(f'A{7 + idx}', [cat.as_list()])
pass



from dataclasses import dataclass
import decimal, json
from datetime import datetime
import gspread, json, google.auth, dataclasses
from googleapiclient.discovery import build
from gspread import Spreadsheet, Worksheet
from datetime import datetime
import wdig.config as config
import pandas as pd
from gspread_formatting import set_column_widths, format_cell_ranges, format_cell_range, get_conditional_format_rules
from gspread_formatting.models import CellFormat, TextFormat
from wdig.queries import transactions_dataframe, merchant_type_summary, get_transactions_grouped_by_budget_tag, BudgetTagGroups


@dataclass
class GoogleWorkspace:
    drive: object
    gspread: gspread.Client
    is_dev: bool


class EnhancedJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, decimal.Decimal):
            return (str(o) for o in [o])  # https://stackoverflow.com/a/1960649
        return super(EnhancedJSONEncoder, self).default(o)

class DecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            return str(o)
        return super(DecimalEncoder, self).default(o)


# budget = {'income': 9600, 'fixed': -1798.74, 'variable': -3108.33, 'mortgage': -2594.98, 'savings': -950, 'donation': -293.33, 'bank fees': -95, 'debt': -340}


# https://developers.google.com/drive/api/v3/about-sdk
# https://developers.google.com/sheets/api/guides/concepts
# https://docs.gspread.org/en/v3.7.0/index.html


_title_cell_format = CellFormat(
    textFormat=TextFormat(
        bold=True,
        foregroundColor={
            "red": 1.0,
            "green": 1.0,
            "blue": 1.0
        },
    ),
    horizontalAlignment="CENTER",
    backgroundColor={
        "red": 1.0,
        "green": 0.0,
        "blue": 0.0
    },
)

_google_workspace = None


def google_connect() -> GoogleWorkspace:
    global _google_workspace
    if _google_workspace is None:
        SCOPES = ["https://www.googleapis.com/auth/drive.metadata.readonly"]
        creds, project = google.auth.load_credentials_from_file(
            config.google_auth_filepath, scopes=SCOPES
        )
        drive_service = build("drive", "v3", credentials=creds)
        gspread_service = gspread.service_account(filename=config.google_auth_filepath)
        _google_workspace = GoogleWorkspace(drive_service, gspread_service, config.development)
    return _google_workspace


def get_or_create_sheet(sheet_title: str, first_worksheet_title: str = 'Sheet1') -> Worksheet:
    google = google_connect()
    if google.is_dev:
        sheet_title += ' DEV'

    sheet = next(iter(google.gspread.openall(title=sheet_title)), None)
    if sheet is None:
        sheet = google.gspread.create(sheet_title, folder_id=config.google_wdig_folder_id)
        worksheet = sheet.get_worksheet(0)
        worksheet.update_title(first_worksheet_title)
    return sheet


def update_empty_sheet():
    sheet = get_or_create_sheet('wdig - test sheet', first_worksheet_title='status')
    worksheet = sheet.get_worksheet(0)
    worksheet.update([['last updated:', datetime.now().strftime('%x'), datetime.now().strftime('%X')]])

def get_transaction_sheet(year: str = '2022') -> Worksheet:
    sheet = get_or_create_sheet(f'wdig - {year}', first_worksheet_title='categories')
    return sheet

def update_transaction_sheet(year: str = '2022') -> None:
    sheet = get_transaction_sheet(year)

    df = transactions_dataframe()
    df = df[df['period_year'] == year]
    df['tran_date'] = pd.to_datetime(df['tran_date']).dt.strftime('%d %b')
    df = df[['period_month', 'tran_date', 'category', 'budget_category', 'budget_tag', 'description', 'amount', 'tran_type', 'account_name', 'bank_name', 'merchant_type', 'tran_id', 'file_name']]

    _update_category_worksheet(df, _create_or_replace_worksheet(sheet, 'categories'))
    _update_transaction_worksheet(df, _create_or_replace_worksheet(sheet, 'transactions'))
    # _update_merchant_type_summary_worksheet(_create_or_replace_worksheet(sheet, 'merchant types'), year)
    _update_budget_worksheet(_create_or_replace_worksheet(sheet, 'budget', rows=100), year)


def _update_merchant_type_summary_worksheet(mt_worksheet: Worksheet, year: str) -> None:
    results = merchant_type_summary(year)
    mt_worksheet.update(results)
    format_cell_ranges(mt_worksheet, [("1", _title_cell_format)])
    mt_worksheet.set_basic_filter()
    set_column_widths(mt_worksheet, [('A', 116), ('B', 122), ('C', 96), ('D', 111)])
    format_cell_range(mt_worksheet, 'D', CellFormat(numberFormat={'type': 'CURRENCY', 'pattern': '"$"#,##0.00'}))
    mt_worksheet.freeze(rows=1)


def _update_transaction_worksheet(df_trans: pd.DataFrame, year_worksheet: Worksheet):
    values = json.loads(json.dumps(df_trans.values.tolist(), default=str))
    year_worksheet.update([df_trans.columns.values.tolist()] + values)

    format_cell_ranges(year_worksheet, [("1", _title_cell_format)])
    year_worksheet.set_basic_filter()
    set_column_widths(year_worksheet, [('A', 116), ('B', 87), ('C', 87), ('D', 122), ('E', 122), ('F', 319), ('G', 75), ('H', 87), ('I', 156), ('J', 100), ('K', 262), ('L', 391)])
    format_cell_range(year_worksheet, 'F', CellFormat(numberFormat={'type': 'CURRENCY', 'pattern': '"$"#,##0.00'}))
    year_worksheet.freeze(rows=1)


def _update_category_worksheet(df_trans: pd.DataFrame, cat_worksheet: Worksheet):
    df_cat = df_trans.groupby(['period_month', 'category']).agg(
        count=pd.NamedAgg(column='amount', aggfunc='count'),
        amount=pd.NamedAgg(column='amount', aggfunc='sum')
    )
    df_cat.reset_index(inplace=True)

    df_cat = df_cat.pivot(
        index=["period_month"], columns="category", values="amount"
    )
    df_cat.reset_index(inplace=True)
    df_cat.fillna(0, inplace=True)
    df_cat['remaining'] = df_cat.sum(axis=1)

    values = json.loads(json.dumps(df_cat.values.tolist(), default=str))
    cat_worksheet.update('A2', [df_cat.columns.values.tolist()] + values)
    format_cell_ranges(cat_worksheet, [("2", _title_cell_format)])


def _update_budget_worksheet(budget_worksheet: Worksheet, year: str):
    budget = {
        'fixed': {
            'mobile usage': -26,
            'disney': -12.99,
            'school fees': -169,
            'apple': -2,
            'casey gym': -87,
            'google storage+youtube': -24.48,
            'swim lessons': 0,
            'life insurance': -21.26,
            'water': -80,
            'southern cross': -127.86,
            'amazon music': -8.5,
            'city council': -246,
            'recycling': -27.16,
            'orcon': -88,
            'minecraft realm': -5.9,
            'home insurance': -142.07,
            'car insurance': -17.51,
            'backup': -19,
            'netflix': -12.99,
            'onedrive': -3,
            'tkd': -190,
            'school special character': -53,
            'kindy': -120,
            'reading eggs': -14.99,
            'power': -250,
        },
        'debt': {
            'cc payments': -340,
            'bank fees': -20,
            'interest': -75,
        },
        'variable': {
            'groceries': -1300,
            'eating out': -433.33,
            'entertainment & misc': -1083.33,
            'petrol': -108.33,
            'at hop': -100.00,
        },
        'donations': {
            'salvation army': -20,
            'kiva': -40,
            'church': -180,
            'city mission': -60,
        },
        'mortgage': {'mortgage': -2594.98},
        'savings': {'savings': -900},
        'unknown': {'unknown': -1085}
    }

    assert len(budget['fixed']) == 25
    assert len(budget['debt']) == 3
    assert len(budget['unknown']) == 1

    actuals = get_transactions_grouped_by_budget_tag(year)

    rows = []
    rows.append(['category', 'tag', 'budget'])
    for m in range(1, datetime.now().month + 1):
        period = f'{datetime(int(year), m, 1).strftime("%m-%b")}'
        rows[0].append(period)

# TODO add category total row

    for bc in budget.keys():
        print(bc)
        for bt in budget[bc].keys():
            row = [bc, bt, float(budget[bc][bt])]
            for m in range(1, datetime.now().month + 1):
                period = f'{datetime(int(year), m, 1).strftime("%m-%b")}'

                def filter_budget_group(group: BudgetTagGroups) -> bool:
                    if group.period == period and group.budget_category == bc and group.budget_tag == bt:
                        return True

                match = list(filter(filter_budget_group, actuals))
                actual = 0
                if len(match) == 1:
                    actual = match[0].amount
                elif len(match) == 0:
                    actual = 0
                else:
                    raise KeyError(f'too many matches for [{bc}] [{bt}]')
                row.append(float(actual))
            rows.append(row)

    budget_worksheet.update(rows)
    format_cell_ranges(budget_worksheet, [("1", _title_cell_format)])
    budget_worksheet.set_basic_filter()
    budget_worksheet.freeze(rows=1)


def _create_or_replace_worksheet(sheet: Spreadsheet, title: str, rows: int = 10, cols: int = 10) -> Worksheet:
    worksheet = _find_worksheet(sheet, title)
    if worksheet:
        sheet.del_worksheet(worksheet)
    worksheet = sheet.add_worksheet(title=title, rows=rows, cols=cols)
    return worksheet


def _find_worksheet(sheet: Spreadsheet, title: str) -> Worksheet:
    worksheets = sheet.worksheets()
    for ws in worksheets:
        if ws.title == title:
            return ws
    return None


if __name__ == '__main__':
    # _update_budget_worksheet(None, '2022')
    update_transaction_sheet()



"""
simplest possible abstraction around object storage
"""

from sys import prefix
from typing import List
import boto3


class ObjectStorage:
    def __init__(self, key_prefix: str) -> None:
        ssm_client = boto3.client('ssm')
        param = ssm_client.get_parameter(Name='/wdig/dev/databucket/name')
        self.data_bucket_name = param['Parameter']['Value']
        self.s3_client = boto3.client('s3')
        self.key_prefix = key_prefix

    def health_check(self):
        self.s3_client.list_objects_v2(Bucket=self.data_bucket_name)

    def test_access(self):
        self.s3_client.put_object(Bucket=self.data_bucket_name, Key=f'test-access.txt', Body=b'hello - this file is for testing access to S3')
        self.s3_client.get_object(Bucket=self.data_bucket_name, Key=f'test-access.txt')

    def save_file(self, key: str, path_to_file: str) -> None:
        self.s3_client.upload_file(path_to_file, self.data_bucket_name, f'{self.key_prefix}/{key}')

    def list_files(self) -> List[str]:
        response = self.s3_client.list_objects_v2(Bucket=self.data_bucket_name, Prefix=f'{self.key_prefix}/')
        return list(map(lambda obj: obj['Key'], response['Contents']))


class BankFileObjectStorage(ObjectStorage):
    def __init__(self) -> None:
        super().__init__('wdig-bank-files')


if __name__ == "__main__":
    os = BankFileObjectStorage()
    os.health_check()
    os.test_access()
    os.list_files()
    pass



from optparse import check_choice
import os
from xml.dom import NotFoundErr
from rich.console import Console
from wdig.google_sheets import google_connect
from wdig.queries import get_transaction_count
import wdig.config as config


console = Console()


class HealthChecks:
    def check_health(self) -> None:
        health_checks = {}

        def check_data_in_folder_exists():
            if not os.path.exists(config.data_in_path):
                raise NotFoundErr()
        health_checks['check datafolder exists'] = check_data_in_folder_exists

        def check_db():
            get_transaction_count()
        health_checks['check database access'] = check_db

        def check_google_api():
            google_connect()
        health_checks['check google api access'] = check_google_api

        def check_data_file_count():
            file_count = 0
            for root, dirnames, filenames in os.walk(config.data_in_path):
                for filename in filenames:
                    if filename.lower().endswith('.csv'):
                        file_count += 1

            if file_count <= 0:
                raise NotFoundErr()
        health_checks['check data file count'] = check_data_file_count

        for key in health_checks.keys():
            console.print(f'[bold]{key}...', end='')
            try:
                health_checks[key]()
                console.print(f"[green] \u2713 PASSED")
            except Exception as ex:
                console.print(f"[red] \u2717 FAILED")
                console.print(f"[red]{ex}")


if __name__ == "__main__":  # pragma: no cover
    hc = HealthChecks()
    hc.check_health()



from dataclasses import dataclass
from typing import List
import pandas as pd
from rich.console import Console
from datetime import timedelta, datetime
from wdig.database import DatabaseAppSession, Transaction
from sqlalchemy import text, func


console = Console()


def query_by_period(months_back: int = 6) -> pd.DataFrame:
    df = transactions_dataframe()
    date_param = (datetime.now() - timedelta(days=months_back*30)).strftime('%Y-%m-%d')
    df = df[df['tran_date'] > date_param]

    df_period = df[['period_month', 'period_year', 'amount']].groupby(['period_year', 'period_month']).agg(
        count=pd.NamedAgg(column='amount', aggfunc='count'),
        amount=pd.NamedAgg(column='amount', aggfunc='sum')
    )
    df_period.reset_index(inplace=True)
    df_period['amount'] = df_period['amount'].round(2)

    return df_period


def get_transaction_count() -> int:
    db_conn = DatabaseAppSession()
    stmt = db_conn.db_session.query(Transaction)
    tran_count = db_conn.db_session.query(Transaction).count()
    return tran_count


def transactions_dataframe() -> pd.DataFrame:
    db_conn = DatabaseAppSession()
    df = pd.read_sql_query(
        "select period_month, period_year, amount, category, budget_category, budget_tag, merchant_type, tag, account_name, description, id, tran_id, tran_type, account_number, tran_date, bank_name, account_format, imported_date, file_name from transaction where is_duplicate is false and category != 'transfer'",
        db_conn.db_engine.engine,
    )
    return df


def is_file_already_loaded(file_name: str) -> bool:
    db_conn = DatabaseAppSession()
    count = (
        db_conn.db_session.query(Transaction)
        .filter(Transaction.file_name == file_name)
        .count()
    )
    return count > 0


def get_unenriched_transaction_count() -> int:
    db_conn = DatabaseAppSession()
    count = (
        db_conn.db_session.query(Transaction)
        .filter(Transaction.category == None)  # noqa: E711
        .count()
    )
    return count


def get_most_recent_tran_date() -> datetime:
    db_conn = DatabaseAppSession()
    tran = db_conn.db_session.query(Transaction).order_by(Transaction.tran_date.desc()).limit(1).first()
    return tran.tran_date


def merchant_type_summary(year: str = '2022') -> List[List]:
    stmt = text('''
select period_month, merchant_type, count(merchant_type), sum(amount)*-1
from transaction
where period_year = :year and category = 'variable' and is_duplicate = false
group by period_month, merchant_type
order by period_month;
                ''')
    stmt = stmt.bindparams(year=year)
    db_conn = DatabaseAppSession()
    results = db_conn.db_engine.engine.execute(stmt).fetchall()
    typed_results = [['period_month', 'merchant_type', 'tran_count', 'total_amount']]
    for row in results:
        typed_results.append([row[0], row[1], row[2], float(row[3])])
    return typed_results


def get_tran_by_id(tran_id: str) -> Transaction:
    db_conn = DatabaseAppSession()
    tran = (
        db_conn.db_session.query(Transaction)
        .filter(Transaction.tran_id == tran_id)
        .first()
    )
    return tran

@dataclass
class BudgetTagGroups:
    budget_category: str
    budget_tag: str
    amount: float
    period: str

def get_transactions_grouped_by_budget_tag(year: str = '2022') -> List[BudgetTagGroups]:
    db_conn = DatabaseAppSession()
    raw_results = (
        db_conn.db_session
        .query(Transaction.budget_category, Transaction.budget_tag, func.sum(Transaction.amount), Transaction.period_month)
        .filter(Transaction.is_duplicate == False)  # noqa E712
        .filter(Transaction.period_year == year)
        .filter(Transaction.category != 'income')
        .filter(Transaction.category != 'transfer')
        .group_by(Transaction.budget_category, Transaction.budget_tag, Transaction.period_month)
        .order_by(Transaction.period_month)
        .all()
    )
    results = []
    for tran in raw_results:
        bg = BudgetTagGroups(tran[0], tran[1], tran[2], tran[3])
        if not bg.budget_category:
            bg.budget_category = 'unknown'
            bg.budget_tag = 'unknown'
        results.append(bg)
    return results


if __name__ == '__main__':
    budget_tags = get_transactions_grouped_by_budget_tag()
    pass



from wdig.database import Transaction


def map_to_mortgage_budget(tran: Transaction) -> None:
    budget_tag = None

    if 'CASEY & MIHEE LOAN PAYMT' in tran.description.upper():
        budget_tag = 'mortgage'

    if budget_tag:
        tran.budget_tag = budget_tag
        tran.budget_category = 'mortgage'
        return True



from wdig.database import Transaction, DatabaseAppSession
from sqlalchemy import func
from rich.console import Console

console = Console()


def dedup():
    db_conn = DatabaseAppSession()
    duplicated_tran_ids = (
        db_conn.db_session.query(Transaction.tran_id)
        .group_by(Transaction.tran_id)
        .having(func.count(Transaction.tran_id) > 1)
    )

    with console.status(f"[bold green]Deduping duplicate transactions...") as status:
        dup_count = 0
        for dup in duplicated_tran_ids:
            dup_trans = (
                db_conn.db_session.query(Transaction)
                .filter(Transaction.tran_id == dup.tran_id)
                .order_by(Transaction.imported_date.desc())
            )

            most_recent_file_name = ""
            for dup_tran in dup_trans:
                if most_recent_file_name == "":
                    most_recent_file_name = dup_tran.file_name

                if dup_tran.file_name != most_recent_file_name:
                    dup_tran.is_duplicate = True
                    dup_count += 1

                db_conn.db_session.commit()
    console.print(f'Deduped duplicate transactions')

    return dup_count


if __name__ == "__main__":
    dedup()



import uuid
from sqlalchemy.sql.expression import and_
from wdig.database import Transaction, DatabaseAppSession
from datetime import datetime as dt
from rich.progress import track
from rich.console import Console
from wdig.enrich.periods import apply_period


console = Console()

# 20000	debit	BNZ Revolving Mortgage	BNZ	d013e510-9570-5eae-7c66-b1201cd726fb
# 26353.52	credit	Kiwibank Cash	Kiwibank	ec169870-b45a-4679-e832-620194f3e76c
# 575000	debit	BNZ Revolving Mortgage	BNZ	707654c7-fc23-f2c5-2d4d-d37aadb10eeb
# variable -595000	debit	BNZ Revolving Mortgage	BNZ	aa31bc69-066a-c6fa-fcc7-d4aa37e76829

_adjustments = [
    {
        "amount": -20000,
        "category": "mortgage",
        "reason": "mortgage transfer for d013e510-9570-5eae-7c66-b1201cd726fb",
        "month": "1",
        "year": 2022,
    },
    {
        "amount": -26353.52,
        "category": "mortgage",
        "reason": "mortgage transfer for ec169870-b45a-4679-e832-620194f3e76c",
        "month": "1",
        "year": 2022,
    },
    {
        "amount": -575000,
        "category": "mortgage",
        "reason": "mortgage transfer for 707654c7-fc23-f2c5-2d4d-d37aadb10eeb",
        "month": "1",
        "year": 2022,
    },
    {
        "amount": 595000,
        "category": "mortgage",
        "reason": "mortgage transfer for aa31bc69-066a-c6fa-fcc7-d4aa37e76829",
        "month": "1",
        "year": 2022,
    },
    {
        "amount": 3078,
        "category": "transfer",
        "reason": "cc payment split over month end",
        "month": "2",
        "year": 2021,
    },
    {
        "amount": -3078,
        "category": "transfer",
        "reason": "cc payment split over month end",
        "month": "3",
        "year": 2021,
    },
    {
        "amount": 3078,
        "category": "transfer",
        "reason": "cc payment split over month end",
        "month": "7",
        "year": 2021,
    },
    {
        "amount": -3685.44,
        "category": "transfer",
        "reason": "cc payment split over month end",
        "month": "8",
        "year": 2021,
    },

]


def apply_adjustments() -> int:
    adjustment_tag_value = "adjustment"

    adjustment_count = 0
    for adj in track(_adjustments, description='Adding adjustments...'):
        db_conn = DatabaseAppSession()

        tran = (
            db_conn.db_session.query(Transaction)
            .filter(
                and_(
                    Transaction.tag == adjustment_tag_value,
                    Transaction.amount == adj["amount"],
                )
            )
            .first()
        )
        if not tran:
            tran = Transaction()
            db_conn.db_session.add(tran)
            adjustment_count += 1

        tran.tran_id = uuid.uuid4().hex
        tran.tran_type = "adjustment"
        tran.tag = adjustment_tag_value
        tran.amount = adj["amount"]
        tran.category = adj["category"]
        tran.account_name = "adjustment"
        tran.account_format = "adjustment"
        tran.account_number = "adjustment"
        tran.bank_name = "adjustment"
        tran.file_name = "adjustment"
        tran.original_columns = {}
        tran.imported_date = dt.now()
        tran.is_duplicate = False
        tran.tran_date = dt(year=int(adj["year"]), month=int(adj["month"]), day=19)
        tran.description = f'adjustment: {adj["reason"]}'
        apply_period(tran)

        db_conn.db_session.commit()

    console.print(f"added {adjustment_count} adjustments")
    return adjustment_count


if __name__ == '__main__':
    apply_adjustments()



from wdig.database import TransactionIterator
from wdig.enrich.categories import apply_category
from wdig.enrich.periods import apply_period
from rich.console import Console
from rich.progress import track
from wdig.queries import get_unenriched_transaction_count, get_transaction_count
from wdig.enrich.clean_data import dedup
from wdig.enrich.adjustments import apply_adjustments
from wdig.enrich.merchant_type import determine_merchant_type


console = Console()


def enrich_all_transactions():
    tran_count = get_transaction_count()
    unenriched_count = get_unenriched_transaction_count()
    console.print(f'found {tran_count} transactions, {unenriched_count} missing enrichment')

    iterator = TransactionIterator(period_month='10-Oct', period_year='2022')
    for trans in track(iterator, total=iterator.total_batches, description='Enriching transactions...'):
        for tran in trans:
            apply_period(tran)
            apply_category(tran)
            tran.merchant_type = determine_merchant_type(tran.description)
    apply_adjustments()
    dedup()



import os
from typing import List
from rich.console import Console
from wdig.parse import ParsedTransaction, parse_row
import csv
from wdig.database import add_parsed_transaction, commit_changes
from wdig.queries import is_file_already_loaded
import wdig.config as config
from wdig.loading.object_storage import BankFileObjectStorage
from wdig.database_v2 import Transaction, database_session


console = Console()


class BankFileLoader():

    def load_new_files(self) -> int:

        file_loaded_count = 0
        for root, dirnames, filenames in os.walk(config.data_in_path):
            for filename in filenames:
                if not filename.lower().endswith('.csv'):
                    continue

                in_filepath = os.path.join(root, filename)
                if self._is_file_already_loaded(filename):
                    continue

                self._save_object(filename, in_filepath)

                with open(in_filepath, 'r') as infile:
                    trans = []
                    for row in csv.DictReader(infile):
                        tran = parse_row(row, filename)
                        self._add_parsed_transaction(tran)
                        trans.append(tran)
                self._commit_changes()
                file_loaded_count += 1
                console.print(f'loaded {len(trans)} in {os.path.basename(in_filepath)}')

        console.print(f'[bold green] found {file_loaded_count} new files')
        return file_loaded_count

    def _save_object(self, filename: str, path_to_file: str) -> None:
        object_storage = BankFileObjectStorage()
        object_storage.save_file(key=filename, path_to_file=path_to_file)

    def _is_file_already_loaded(self, filename):
        # query it here?
        return False


class BankFileLoaderV2():
    def get_files():
        # transactions by file_name, count(),
        pass

    def load_file(self, path_to_file: str, force: bool = False) -> False:  # return load status?
        filename = os.path.basename(path_to_file)
        with open(path_to_file, 'r') as file:
            trans = []
            for row in csv.DictReader(file):
                parsed_tran = parse_row(row, filename)
                tran = self._to_transaction(parsed_tran)
                trans.append(tran)
            with database_session() as db:
                db.add_all(trans)
                db.commit()

    def load_new_files(self) -> List[str]:
        new_files = []
        for root, dirnames, filenames in os.walk(config.data_in_path):
            for filename in filenames:
                if not filename.lower().endswith('.csv'):
                    continue

                # TODO if not already loaded
                self.load_file()

                new_files.append(filename)

    def _to_transaction(self, parsed_tran: ParsedTransaction) -> Transaction:
        tran = Transaction()
        tran.tran_id = parsed_tran.unique_id
        tran.tran_type = parsed_tran.tran_type
        tran.amount = parsed_tran.amount
        tran.tran_date = parsed_tran.tran_date
        tran.description = parsed_tran.description
        tran.account_number = parsed_tran.account_number
        tran.account_name = parsed_tran.account_name
        tran.bank_name = parsed_tran.bank_name
        tran.account_format = parsed_tran.account_format
        tran.original_columns = parsed_tran.original
        tran.imported_date = parsed_tran.imported_date
        tran.file_name = parsed_tran.file_name
        tran.is_duplicate = False


# TODO the abstraction (resource access) should be testable!
class BankFileLoaderTestable(BankFileLoader):
    def __init__(self, is_file_already_loaded: bool) -> None:
        super().__init__()
        self.is_file_already_loaded = is_file_already_loaded

    def _is_file_already_loaded(self, filename):
        return self.is_file_already_loaded

    def _save_object(self, filename: str, path_to_file: str) -> None:
        pass

    def _add_parsed_transaction(self, tran):
        pass

    def _commit_changes(self):
        pass


if __name__ == '__main__':
    bfl = BankFileLoaderV2
    bfl.load_new_files()



"""
using TOML for config files https://toml.io/en/
"""

import os
import os.path as path
import toml
import json


class Config():
    _values = {}

    def __init__(self) -> None:
        self._path_to_config_file = path.expanduser('~/.wdig/config.toml')
        self.load()

    def _set_default_values(self):
        self._values = {}
        self._values['path_to_sqlite_db'] = '~/.wdig/wdig.sqlite.db'
        self._values['path_to_bank_files'] = ''

    def load(self):
        if not path.exists(self.path_to_config_file):
            self.reset_to_default()

        with open(self.path_to_config_file, 'r') as file:
            self._values = toml.load(file)

    def save(self):
        os.makedirs(path.dirname(self.path_to_config_file), exist_ok=True)
        with open(self.path_to_config_file, 'w+') as file:
            toml.dump(self._values, file)

    def reset_to_default(self) -> None:
        if path.exists(self.path_to_config_file):
            os.remove(self.path_to_config_file)
        self._set_default_values()
        self.save()

    def to_toml(self) -> str:
        return toml.dumps(self._values)

    def to_json(self) -> str:
        return json.dumps(self._values)

    def to_dict(self) -> dict:
        return dict.copy(self._values)

    @property
    def path_to_config_file(self) -> str:
        return self._path_to_config_file

    @property
    def path_to_sqlite_db(self) -> str:
        return path.expanduser(self._values['path_to_sqlite_db'])

    @property
    def path_to_bank_files(self) -> str:
        return path.expanduser(self._values['path_to_bank_files'])

    @path_to_bank_files.setter
    def path_to_bank_files(self, value: str):
        self._values['path_to_bank_files'] = value
        self.save()


config = Config()

if __name__ == '__main__':
    config.reset_to_default()
    print(config.path_to_config_file)


development = ('WDIG-LIVE' not in os.environ.keys())

app_name = "wdig"
if development:
    app_name += '-dev'

version = 'x.y.z'  # how to get this from docker tags?


# google
google_auth_filepath = '.secrets/google-auth.json'
google_wdig_folder_id = "1XjS2bvm56rwv0D3KwaUAAmy3E5QCY0kZ"


# data
data_in_path = 'data-in'


# database
db_username = 'postgres'
db_password = os.environ.get('WDIG_DB_PASSWORD')
db_hostname = os.environ.get('WDIG_DB_HOST', 'localhost')
db_hostport = '5432'
# pick db name based on config.development
db_dbname = 'wdig_dev'
if not development:
    db_dbname = 'wdig'



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wdig', 'wdig.enrich', 'wdig.loading', 'wdig.outputs']

package_data = \
{'': ['*']}

install_requires = \
['boto3>=1.26.27,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'numpy>=1.23.5,<2.0.0',
 'pandas>=1.5.2,<2.0.0',
 'rich>=12.6.0,<13.0.0',
 'sqlalchemy>=1.4.44,<2.0.0',
 'tinydb>=4.7.0,<5.0.0',
 'toml>=0.10.2,<0.11.0']

setup_kwargs = {
    'name': 'wdig-cli',
    'version': '0.1.1',
    'description': 'Where Did It Go?',
    'long_description': '# wdig\n\n[![wdig](https://github.com/jebu4/wdig/actions/workflows/wdig.yml/badge.svg)](https://github.com/jebu4/wdig/actions/workflows/wdig.yml)\n\nwhere did it go?\n\n## tasks\n\n### dev ci\n\n```bash\nnox\n```\n',
    'author': 'Casey Burns',
    'author_email': 'CaseyBurns@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import click
from datetime import timedelta, datetime
from wdig.loading import BankFileLoaderV2
from rich.console import Console


console = Console()


@click.group('files')
def cli_files():
    pass

@click.command()
def list():
    pass

@click.command()
def new():
    bfl = BankFileLoaderV2()
    new_files = bfl.load_new_files()

    if len(new_files) == 0:
        last_tran_date = bfl.get_date_for_new_files()
        console.print(f'go and download files from "{last_tran_date.strftime("%-d %b %y")}" to today')
        console.print('these accounts; Kiwibank (cash, savings, onlinebills) bnz (revolving, cash, fixed) anz (cash, creditcard)')
        console.print(f'put them in a folder named {datetime.now().strftime("%Y-%m-%d")}')


cli_files.add_command(list)
cli_files.add_command(new)



from dataclasses import dataclass
import os
from typing import List
from rich.console import Console
from wdig.parse import ParsedTransaction, parse_row
import csv
from wdig.config import config
from wdig.database_v2 import Transaction, database_session
from datetime import datetime, timedelta
from sqlalchemy import func
from wdig.user_notifier import UserNotifier, NotificationMessage


console = Console()
notifier = UserNotifier()


@dataclass
class BankFileInfo():
    file_name: str
    transaction_count: int


class BankFileLoaderV2():
    def get_files(self):
        with database_session() as db:
            results = (
                db
                .query(Transaction.file_name, func.count(Transaction.file_name))
                .group_by(Transaction.file_name)
                .order_by(Transaction.file_name)
                .all()
            )
        # TODO map results to dataclass - nice way to do this?

    def get_date_for_new_files(self) -> datetime:
        with database_session() as db:
            tran = db.query(Transaction).order_by(Transaction.tran_date.desc()).limit(1).first()
            if tran:
                return tran.tran_type + timedelta(days=-2)
            else:
                return datetime.now() + timedelta(days=-90)

    def load_file(self, path_to_file: str, force: bool = False) -> int:
        """returns number of trans loaded"""
        filename = os.path.basename(path_to_file)
        # TODO if not already loaded, or force
        with open(path_to_file, 'r') as file:
            trans = []
            for row in csv.DictReader(file):
                parsed_tran = parse_row(row, filename)
                tran = self._to_transaction(parsed_tran)
                trans.append(tran)
            with database_session() as db:
                db.add_all(trans)
                db.commit()
        notifier.send(NotificationMessage(f'loaded {len(trans)} transactions from {filename}'))
        return len(trans)

    def load_new_files(self) -> List[str]:
        new_files = []
        for root, dirnames, filenames in os.walk(config.path_to_bank_files):
            for filename in filenames:
                if not filename.lower().endswith('.csv'):
                    continue
                # TODO check if file already loaded
                new_files.append(filename)
        notifier.send(NotificationMessage(f'found {len(new_files)} new bank files to load'))

        total_transactions = 0
        for filename in new_files:
            tran_count = self.load_file(os.path.join(root, filename))
            total_transactions += tran_count
        notifier.send(NotificationMessage(f'successfully loaded {total_transactions} transactions from {len(new_files)} new files'))

        return new_files

    def _to_transaction(self, parsed_tran: ParsedTransaction) -> Transaction:
        tran = Transaction()
        tran.tran_id = parsed_tran.unique_id
        tran.tran_type = parsed_tran.tran_type
        tran.amount = parsed_tran.amount
        tran.tran_date = parsed_tran.tran_date
        tran.description = parsed_tran.description
        tran.account_number = parsed_tran.account_number
        tran.account_name = parsed_tran.account_name
        tran.bank_name = parsed_tran.bank_name
        tran.account_format = parsed_tran.account_format
        tran.original_columns = parsed_tran.original
        tran.imported_date = parsed_tran.imported_date
        tran.file_name = parsed_tran.file_name
        tran.is_duplicate = False
        return tran


if __name__ == '__main__':
    bfl = BankFileLoaderV2()
    # new_files = bfl.load_new_files()
    bfl.get_files()
